@@ -1,8 +1,8 @@
 /**************************************************************************
  *   color.c  --  This file is part of GNU nano.                          *
  *                                                                        *
- *   Copyright (C) 2001-2011, 2013-2018 Free Software Foundation, Inc.    *
- *   Copyright (C) 2014-2017 Benno Schulenberg                            *
+ *   Copyright (C) 2001-2011, 2013-2020 Free Software Foundation, Inc.    *
+ *   Copyright (C) 2014-2017, 2020 Benno Schulenberg                      *
  *                                                                        *
  *   GNU nano is free software: you can redistribute it and/or modify     *
  *   it under the terms of the GNU General Public License as published    *
@@ -19,119 +19,93 @@
  *                                                                        *
  **************************************************************************/
 
-#include "proto.h"
+#include "prototypes.h"
+
+#ifdef ENABLE_COLOR
 
 #include <errno.h>
 #ifdef HAVE_MAGIC_H
 #include <magic.h>
 #endif
 #include <string.h>
-#include <unistd.h>
 
-#ifdef ENABLE_COLOR
+static bool defaults_allowed = FALSE;
+		/* Whether ncurses accepts -1 to mean "default color". */
 
-/* For early versions of ncurses-6.0, use an additional A_PROTECT attribute
- * for all colors, in order to work around an ncurses miscoloring bug. */
-#if defined(NCURSES_VERSION_MAJOR) && (NCURSES_VERSION_MAJOR == 6) && \
-		(NCURSES_VERSION_MINOR == 0) && (NCURSES_VERSION_PATCH < 20151017)
-#define A_BANDAID  A_PROTECT
-#else
-#define A_BANDAID  A_NORMAL
-#endif
-
-/* Initialize the colors for nano's interface, and assign pair numbers
- * for the colors in each syntax. */
-void set_colorpairs(void)
+/* Initialize the color pairs for nano's interface. */
+void set_interface_colorpairs(void)
 {
-	const syntaxtype *sint;
-	bool using_defaults = FALSE;
-	size_t i;
-
-	/* Tell ncurses to enable colors. */
-	start_color();
-
 #ifdef HAVE_USE_DEFAULT_COLORS
-	/* Allow using the default colors, if available. */
-	using_defaults = (use_default_colors() != ERR);
+	/* Ask ncurses to allow -1 to mean "default color". */
+	defaults_allowed = (use_default_colors() == OK);
 #endif
 
 	/* Initialize the color pairs for nano's interface elements. */
-	for (i = 0; i < NUMBER_OF_ELEMENTS; i++) {
-		colortype *combo = color_combo[i];
+	for (size_t index = 0; index < NUMBER_OF_ELEMENTS; index++) {
+		colortype *combo = color_combo[index];
 
 		if (combo != NULL) {
-			if (combo->fg == USE_THE_DEFAULT && !using_defaults)
-				combo->fg = COLOR_WHITE;
-			if (combo->bg == USE_THE_DEFAULT && !using_defaults)
-				combo->bg = COLOR_BLACK;
-			init_pair(i + 1, combo->fg, combo->bg);
-			interface_color_pair[i] = COLOR_PAIR(i + 1) | A_BANDAID |
-										combo->attributes;
+			if (!defaults_allowed) {
+				if (combo->fg == THE_DEFAULT)
+					combo->fg = COLOR_WHITE;
+				if (combo->bg == THE_DEFAULT)
+					combo->bg = COLOR_BLACK;
+			}
+			init_pair(index + 1, combo->fg, combo->bg);
+			interface_color_pair[index] = COLOR_PAIR(index + 1) | combo->attributes;
 		} else {
-			if (i == FUNCTION_TAG)
-				interface_color_pair[i] = A_NORMAL;
-			else if (i == ERROR_MESSAGE) {
-				init_pair(i + 1, COLOR_WHITE, COLOR_RED);
-				interface_color_pair[i] = COLOR_PAIR(i + 1) | A_BOLD | A_BANDAID;
+			if (index == FUNCTION_TAG || index == SCROLL_BAR)
+				interface_color_pair[index] = A_NORMAL;
+			else if (index == GUIDE_STRIPE)
+				interface_color_pair[index] = A_REVERSE;
+			else if (index == ERROR_MESSAGE) {
+				init_pair(index + 1, COLOR_WHITE, COLOR_RED);
+				interface_color_pair[index] = COLOR_PAIR(index + 1) | A_BOLD;
 			} else
-				interface_color_pair[i] = hilite_attribute;
+				interface_color_pair[index] = hilite_attribute;
 		}
 
-		free(color_combo[i]);
-	}
-
-	/* For each syntax, go through its list of colors and assign each
-	 * its pair number, giving identical color pairs the same number. */
-	for (sint = syntaxes; sint != NULL; sint = sint->next) {
-		colortype *ink;
-		int new_number = NUMBER_OF_ELEMENTS + 1;
-
-		for (ink = sint->color; ink != NULL; ink = ink->next) {
-			const colortype *beforenow = sint->color;
-
-			while (beforenow != ink && (beforenow->fg != ink->fg ||
-										beforenow->bg != ink->bg))
-				beforenow = beforenow->next;
-
-			if (beforenow != ink)
-				ink->pairnum = beforenow->pairnum;
-			else
-				ink->pairnum = new_number++;
-
-			ink->attributes |= COLOR_PAIR(ink->pairnum) | A_BANDAID;
-		}
+		free(color_combo[index]);
 	}
 }
 
-/* Initialize the color information. */
-void color_init(void)
+/* Assign a pair number to each of the foreground/background color combinations
+ * in the given syntax, giving identical combinations the same number. */
+void set_syntax_colorpairs(syntaxtype *sntx)
 {
-	const colortype *ink;
-	bool using_defaults = FALSE;
-	short foreground, background;
+	short number = NUMBER_OF_ELEMENTS;
+	colortype *older;
 
-	/* If the terminal is not capable of colors, forget it. */
-	if (!has_colors())
-		return;
+	for (colortype *ink = sntx->color; ink != NULL; ink = ink->next) {
+		if (!defaults_allowed) {
+			if (ink->fg == THE_DEFAULT)
+				ink->fg = COLOR_WHITE;
+			if (ink->bg == THE_DEFAULT)
+				ink->bg = COLOR_BLACK;
+		}
 
-#ifdef HAVE_USE_DEFAULT_COLORS
-	/* Allow using the default colors, if available. */
-	using_defaults = (use_default_colors() != ERR);
-#endif
+		older = sntx->color;
 
-	/* For each coloring expression, initialize the color pair. */
-	for (ink = openfile->colorstrings; ink != NULL; ink = ink->next) {
-		foreground = ink->fg;
-		background = ink->bg;
+		while (older != ink && (older->fg != ink->fg || older->bg != ink->bg))
+			older = older->next;
 
-		if (foreground == USE_THE_DEFAULT && !using_defaults)
-			foreground = COLOR_WHITE;
+		ink->pairnum = (older != ink) ? older->pairnum : ++number;
 
-		if (background == USE_THE_DEFAULT && !using_defaults)
-			background = COLOR_BLACK;
-
-		init_pair(ink->pairnum, foreground, background);
+		ink->attributes |= COLOR_PAIR(ink->pairnum);
 	}
+}
+
+/* Initialize the color pairs for the current syntax. */
+void prepare_palette(void)
+{
+	short number = NUMBER_OF_ELEMENTS;
+
+	/* For each unique pair number, tell ncurses the combination of colors. */
+	for (colortype *ink = openfile->syntax->color; ink != NULL; ink = ink->next)
+		if (ink->pairnum > number) {
+			init_pair(ink->pairnum, ink->fg, ink->bg);
+			number = ink->pairnum;
+		}
 
 	have_palette = TRUE;
 }
@@ -144,7 +118,7 @@ bool found_in_list(regexlisttype *head, const char *shibboleth)
 	regex_t rgx;
 
 	for (item = head; item != NULL; item = item->next) {
-		regcomp(&rgx, fixbounds(item->full_regex), NANO_REG_EXTENDED);
+		regcomp(&rgx, item->full_regex, NANO_REG_EXTENDED);
 
 		if (regexec(&rgx, shibboleth, 0, NULL, 0) == 0) {
 			regfree(&rgx);
@@ -157,11 +131,11 @@ bool found_in_list(regexlisttype *head, const char *shibboleth)
 	return FALSE;
 }
 
-/* Update the color information based on the current filename and content. */
-void color_update(void)
+/* Find a syntax that applies to the current buffer, based upon filename
+ * or buffer content, and load and prime this syntax when needed. */
+void find_and_prime_applicable_syntax(void)
 {
-	syntaxtype *sint = NULL;
-	colortype *ink;
+	syntaxtype *sntx = NULL;
 
 	/* If the rcfiles were not read, or contained no syntaxes, get out. */
 	if (syntaxes == NULL)
@@ -173,56 +147,39 @@ void color_update(void)
 		if (strcmp(syntaxstr, "none") == 0)
 			return;
 
-		for (sint = syntaxes; sint != NULL; sint = sint->next) {
-			if (strcmp(sint->name, syntaxstr) == 0)
+		for (sntx = syntaxes; sntx != NULL; sntx = sntx->next)
+			if (strcmp(sntx->name, syntaxstr) == 0)
 				break;
-		}
 
-		if (sint == NULL && !inhelp)
+		if (sntx == NULL && !inhelp)
 			statusline(ALERT, _("Unknown syntax name: %s"), syntaxstr);
 	}
 
 	/* If no syntax-override string was specified, or it didn't match,
 	 * try finding a syntax based on the filename (extension). */
-	if (sint == NULL && !inhelp) {
-		char *reserved = charalloc(PATH_MAX + 1);
-		char *currentdir = getcwd(reserved, PATH_MAX + 1);
-		char *joinednames = charalloc(PATH_MAX + 1);
-		char *fullname = NULL;
-
-		if (currentdir == NULL)
-			free(reserved);
-		else {
-			/* Concatenate the current working directory with the
-			 * specified filename, and canonicalize the result. */
-			sprintf(joinednames, "%s/%s", currentdir, openfile->filename);
-			fullname = get_full_path(joinednames);
-			free(currentdir);
-		}
+	if (sntx == NULL && !inhelp) {
+		char *fullname = get_full_path(openfile->filename);
 
 		if (fullname == NULL)
 			fullname = mallocstrcpy(fullname, openfile->filename);
 
-		for (sint = syntaxes; sint != NULL; sint = sint->next) {
-			if (found_in_list(sint->extensions, fullname))
+		for (sntx = syntaxes; sntx != NULL; sntx = sntx->next)
+			if (found_in_list(sntx->extensions, fullname))
 				break;
-		}
 
-		free(joinednames);
 		free(fullname);
 	}
 
 	/* If the filename didn't match anything, try the first line. */
-	if (sint == NULL && !inhelp) {
-		for (sint = syntaxes; sint != NULL; sint = sint->next) {
-			if (found_in_list(sint->headers, openfile->fileage->data))
+	if (sntx == NULL && !inhelp) {
+		for (sntx = syntaxes; sntx != NULL; sntx = sntx->next)
+			if (found_in_list(sntx->headers, openfile->filetop->data))
 				break;
-		}
 	}
 
 #ifdef HAVE_LIBMAGIC
-	/* If we still don't have an answer, try using magic. */
-	if (sint == NULL && !inhelp) {
+	/* If we still don't have an answer, try using magic (when requested). */
+	if (sntx == NULL && !inhelp && ISSET(USE_MAGIC)) {
 		struct stat fileinfo;
 		magic_t cookie = NULL;
 		const char *magicstring = NULL;
@@ -246,10 +203,9 @@ void color_update(void)
 
 		/* Now try and find a syntax that matches the magic string. */
 		if (magicstring != NULL) {
-			for (sint = syntaxes; sint != NULL; sint = sint->next) {
-				if (found_in_list(sint->magics, magicstring))
+			for (sntx = syntaxes; sntx != NULL; sntx = sntx->next)
+				if (found_in_list(sntx->magics, magicstring))
 					break;
-			}
 		}
 
 		if (stat(openfile->filename, &fileinfo) == 0)
@@ -258,34 +214,33 @@ void color_update(void)
 #endif /* HAVE_LIBMAGIC */
 
 	/* If nothing at all matched, see if there is a default syntax. */
-	if (sint == NULL && !inhelp) {
-		for (sint = syntaxes; sint != NULL; sint = sint->next) {
-			if (strcmp(sint->name, "default") == 0)
+	if (sntx == NULL && !inhelp) {
+		for (sntx = syntaxes; sntx != NULL; sntx = sntx->next)
+			if (strcmp(sntx->name, "default") == 0)
 				break;
-		}
 	}
 
-	openfile->syntax = sint;
-	openfile->colorstrings = (sint == NULL ? NULL : sint->color);
-
-	/* If a syntax was found, compile its specified regexes (which have
-	 * already been checked for validity when they were read in). */
-	for (ink = openfile->colorstrings; ink != NULL; ink = ink->next) {
-		if (ink->start == NULL) {
-			ink->start = (regex_t *)nmalloc(sizeof(regex_t));
-			regcomp(ink->start, fixbounds(ink->start_regex), ink->rex_flags);
-		}
-
-		if (ink->end_regex != NULL && ink->end == NULL) {
-			ink->end = (regex_t *)nmalloc(sizeof(regex_t));
-			regcomp(ink->end, fixbounds(ink->end_regex), ink->rex_flags);
-		}
+	/* When the syntax isn't loaded yet, parse it and initialize its colors. */
+	if (sntx != NULL && sntx->filename != NULL) {
+		parse_one_include(sntx->filename, sntx);
+		set_syntax_colorpairs(sntx);
 	}
+
+	openfile->syntax = sntx;
+}
+
+/* Allocate and initialize (for the given line) the cache for multiline info. */
+void set_up_multicache(linestruct *line)
+{
+	line->multidata = nmalloc(openfile->syntax->nmultis * sizeof(short));
+
+	for (short index = 0; index < openfile->syntax->nmultis; index++)
+		line->multidata[index] = -1;
 }
 
 /* Determine whether the matches of multiline regexes are still the same,
  * and if not, schedule a screen refresh, so things will be repainted. */
-void check_the_multis(filestruct *line)
+void check_the_multis(linestruct *line)
 {
 	const colortype *ink;
 	bool astart, anend;
@@ -296,13 +251,14 @@ void check_the_multis(filestruct *line)
 	if (openfile->syntax == NULL || openfile->syntax->nmultis == 0)
 		return;
 
-	for (ink = openfile->colorstrings; ink != NULL; ink = ink->next) {
+	if (line->multidata == NULL)
+		set_up_multicache(line);
+
+	for (ink = openfile->syntax->color; ink != NULL; ink = ink->next) {
 		/* If it's not a multiline regex, skip. */
 		if (ink->end == NULL)
 			continue;
 
-		alloc_multidata_if_needed(line);
-
 		astart = (regexec(ink->start, line->data, 1, &startmatch, 0) == 0);
 		afterstart = line->data + (astart ? startmatch.rm_eo : 0);
 		anend = (regexec(ink->end, afterstart, 1, &endmatch, 0) == 0);
@@ -329,43 +285,30 @@ void check_the_multis(filestruct *line)
 	}
 }
 
-/* Allocate (for one line) the cache space for multiline color regexes. */
-void alloc_multidata_if_needed(filestruct *fileptr)
-{
-	int i;
-
-	if (fileptr->multidata == NULL) {
-		fileptr->multidata = (short *)nmalloc(openfile->syntax->nmultis * sizeof(short));
-
-		for (i = 0; i < openfile->syntax->nmultis; i++)
-			fileptr->multidata[i] = -1;
-	}
-}
-
 /* Precalculate the multi-line start and end regex info so we can
  * speed up rendering (with any hope at all...). */
 void precalc_multicolorinfo(void)
 {
 	const colortype *ink;
 	regmatch_t startmatch, endmatch;
-	filestruct *line, *tailline;
+	linestruct *line, *tailline;
 
-	if (openfile->colorstrings == NULL || ISSET(NO_COLOR_SYNTAX))
+	if (openfile->syntax == NULL || openfile->syntax->nmultis == 0 ||
+					openfile->filetop->multidata || ISSET(NO_SYNTAX))
 		return;
 
-#ifdef DEBUG
-	fprintf(stderr, "Precalculating the multiline color info...\n");
-#endif
+	/* For each line, allocate cache space for the multiline-regex info. */
+	for (line = openfile->filetop; line != NULL; line = line->next)
+		set_up_multicache(line);
 
-	for (ink = openfile->colorstrings; ink != NULL; ink = ink->next) {
+	for (ink = openfile->syntax->color; ink != NULL; ink = ink->next) {
 		/* If this is not a multi-line regex, skip it. */
 		if (ink->end == NULL)
 			continue;
 
-		for (line = openfile->fileage; line != NULL; line = line->next) {
+		for (line = openfile->filetop; line != NULL; line = line->next) {
 			int index = 0;
 
-			alloc_multidata_if_needed(line);
 			/* Assume nothing applies until proven otherwise below. */
 			line->multidata[ink->id] = CNONE;
 
@@ -378,14 +321,14 @@ void precalc_multicolorinfo(void)
 			/* When the line contains a start match, look for an end, and if
 			 * found, mark all the lines that are affected. */
 			while (regexec(ink->start, line->data + index, 1,
-						&startmatch, (index == 0) ? 0 : REG_NOTBOL) == 0) {
+							&startmatch, (index == 0) ? 0 : REG_NOTBOL) == 0) {
 				/* Begin looking for an end match after the start match. */
 				index += startmatch.rm_eo;
 
 				/* If there is an end match on this line, mark the line, but
 				 * continue looking for other starts after it. */
 				if (regexec(ink->end, line->data + index, 1,
-						&endmatch, (index == 0) ? 0 : REG_NOTBOL) == 0) {
+							&endmatch, (index == 0) ? 0 : REG_NOTBOL) == 0) {
 					line->multidata[ink->id] = CSTARTENDHERE;
 					index += endmatch.rm_eo;
 					/* If both start and end are mere anchors, step ahead. */
@@ -394,7 +337,7 @@ void precalc_multicolorinfo(void)
 						/* When at end-of-line, we're done. */
 						if (line->data[index] == '\0')
 							break;
-						index = move_mbright(line->data, index);
+						index = step_right(line->data, index);
 					}
 					continue;
 				}
@@ -417,12 +360,9 @@ void precalc_multicolorinfo(void)
 				 * the lines in between and the end properly. */
 				line->multidata[ink->id] = CENDAFTER;
 
-				for (line = line->next; line != tailline; line = line->next) {
-					alloc_multidata_if_needed(line);
+				for (line = line->next; line != tailline; line = line->next)
 					line->multidata[ink->id] = CWHOLELINE;
-				}
 
-				alloc_multidata_if_needed(tailline);
 				tailline->multidata[ink->id] = CBEGINBEFORE;
 
 				/* Begin looking for a new start after the end match. */
