@@ -477,7 +477,8 @@ size_t shown_entries_for(int menu)
 	}
 
 	/* When --saveonexit is not used, widen the grid of the WriteOut menu. */
-	if (menu == MWRITEFILE && first_sc_for(menu, discard_buffer) == NULL)
+	if (menu == MWRITEFILE && item == NULL &&
+						first_sc_for(menu, discard_buffer) == NULL)
 		count--;
 
 	return count;
@@ -529,6 +530,7 @@ functionptrtype interpret(int *keycode)
 			return do_findnext;
 
 		switch (tolower(*keycode)) {
+			case 'b':
 			case '-':
 				return do_page_up;
 			case ' ':
